@@ -1,8 +1,10 @@
 #ifndef _EXTERN_H_
 #define _EXTERN_H_
 
-#include <netinet/in.h>
 #include <stdio.h>
+#include <signal.h>
+#include <time.h>
+#include <netinet/in.h>
 
 #include "logmessage.h"
 #include "xmalloc.h"
@@ -23,6 +25,7 @@ struct ifmode {
 };
 
 struct tempmode {
+	int	in_use;
 	unsigned char have1, have2, have3;
 };
 
@@ -35,6 +38,9 @@ union wdog_options {
 
 struct list {
 	char *name;
+	int version;
+	time_t last_time;
+	int repair_count;
 	union wdog_options parameter;
 	struct list *next;
 };
@@ -69,13 +75,16 @@ extern int minalloc;
 extern int maxtemp;
 extern int pingcount;
 extern int temp_poweroff;
+extern int sigterm_delay;
+extern int repair_max;
 
 extern char *devname;
 extern char *admin;
 
-extern time_t	test_timeout;		/* test-binary time out value. */
-extern time_t	repair_timeout;		/* repair-binary time out value. */
+extern int	test_timeout;		/* test-binary time out value. */
+extern int	repair_timeout;		/* repair-binary time out value. */
 extern int	dev_timeout;		/* Watchdog hardware time-out. */
+extern int	retry_timeout;		/* Retry on non-critical errors. */
 
 extern char *logdir;
 
@@ -91,8 +100,7 @@ extern struct list *pidfile_list;
 extern struct list *iface_list;
 extern struct list *temp_list;
 
-extern char *tbinary;
-extern char *rbinary;
+extern char *repair_bin;
 
 /* = Not (yet) from config file. = */
 
@@ -127,6 +135,7 @@ int set_watchdog_timeout(int timeout);
 int keep_alive(void);
 int get_watchdog_fd(void);
 int close_watchdog(void);
+void safe_sleep(int sec);
 
 /** load.c **/
 int open_loadcheck(void);
@@ -143,7 +152,8 @@ int check_temp(struct list *act);
 int close_tempcheck(void);
 
 /** test_binary.c **/
-int check_bin(char *, time_t, int);
+int check_bin(char *, int, int);
+void free_process(void);
 
 /** pidfile.c **/
 int check_pidfile(struct list *);
@@ -160,7 +170,7 @@ int check_allocatable(void);
 /** shutdown.c **/
 void do_shutdown(int errorcode);
 void sigterm_handler(int arg);
-void terminate(void) GCC_NORETURN;
+void terminate(int ecode) GCC_NORETURN;
 
 /** heartbeat.c **/
 int open_heartbeat(void);
@@ -174,8 +184,16 @@ void unlock_our_memory(void);
 /** daemon-pid.c **/
 int write_pid_file(const char *fname);
 int remove_pid_file(void);
+int wd_daemon(int nochdir, int noclose);
 
 /** configfile.c **/
 void read_config(char *configfile);
 
+/** errorcodes.c **/
+const char *wd_strerror(int err);
+
+/** sigterm.c **/
+extern volatile sig_atomic_t _running;
+void sigterm_handler(int arg);
+
 #endif /*_EXTERN_H_*/
