@@ -12,32 +12,26 @@
 int check_file_table(void)
 {
 	int fd;
+	const char fname[] = "/proc/uptime"; /* Any file will do, so long as it ALWAYS exists. */
+	int err = ENOERR;
 
 	/* open a file */
-	fd = open("/proc/uptime", O_RDONLY);
+	fd = open(fname, O_RDONLY);
 	if (fd == -1) {
-		int err = errno;
+		err = errno;
+		log_message(LOG_ERR, "cannot open %s (errno = %d = '%s')", fname, err, strerror(err));
 
 		if (err == ENFILE) {
 			/* we need a reboot if ENFILE is returned (file table overflow) */
 			log_message(LOG_ERR, "file table overflow detected!");
-			return (ENFILE);
-		} else {
-			errno = err;
-			log_message(LOG_ERR, "cannot open /proc/uptime (errno = %d = '%s')", err, strerror(err));
-
-			if (softboot)
-				return (err);
+			return (EREBOOT);
 		}
 	} else {
 		if (close(fd) < 0) {
-			int err = errno;
-			log_message(LOG_ERR, "close /proc/uptime gave errno = %d = '%s'", err, strerror(err));
-
-			if (softboot)
-				return (err);
+			err = errno;
+			log_message(LOG_ERR, "close %s gave errno = %d = '%s'", fname, err, strerror(err));
 		}
 	}
 
-	return (ENOERR);
+	return (err);
 }
