@@ -11,7 +11,6 @@
 #include <mntent.h>
 #include <netdb.h>
 #include <paths.h>
-#include <setjmp.h>
 #include <signal.h>
 #include <string.h>
 #include <stdlib.h>
@@ -25,9 +24,9 @@
 
 #include "watch_err.h"
 #include "extern.h"
+#include "ext2_mnt.h"
 
 #if defined __GLIBC__
-#include "ext2_mnt.h"
 #include <sys/quota.h>
 #include <sys/swap.h>
 #include <sys/reboot.h>
@@ -57,11 +56,6 @@ extern int mount_one(char *, char *, char *, char *, int, int);
 static struct mntent rootfs;
 #endif
 
-extern volatile sig_atomic_t _running;	/* From watchdog.c */
-extern int dev_timeout;			/* From watchdog.c */
-
-jmp_buf ret2dog;
-
 /* Info about a process. */
 typedef struct _proc_ {
 	pid_t pid;		/* Process ID.                    */
@@ -69,29 +63,6 @@ typedef struct _proc_ {
 	struct _proc_ *next;	/* Pointer to next struct.        */
 } PROC;
 
-/* A version of sleep() that keeps the watchdog timer alive. */
-static void safe_sleep(int nsec)
-{
-	int i;
-
-	keep_alive();
-	for (i=0; i<nsec; i++) {
-		sleep(1);
-		keep_alive();
-	}
-}
-
-/* write a log entry on exit */
-static void log_end(void)
-{
-	/* Log the closing message */
-	log_message(LOG_INFO, "stopping daemon (%d.%d)", MAJOR_VERSION, MINOR_VERSION);
-	closelog();
-	safe_sleep(5);		/* make sure log is written */
-
-	return;
-}
-
 /* close the device and check for error */
 static void close_all(void)
 {
@@ -100,24 +71,19 @@ static void close_all(void)
 	close_memcheck();
 	close_tempcheck();
 	close_heartbeat();
-}
-
-/*
- * Used to tell the main() code to exit gracefully.
- */
-void sigterm_handler(int arg)
-{
-	_running = 0;
+	free_process();		/* What check_bin() was waiting to report. */
 }
 
 /* on exit we close the device and log that we stop */
-void terminate(void)
+void terminate(int ecode)
 {
+	log_message(LOG_NOTICE, "stopping daemon (%d.%d)", MAJOR_VERSION, MINOR_VERSION);
 	unlock_our_memory();
 	close_all();
 	remove_pid_file();
-	log_end();
-	exit(0);
+	close_logging();
+	usleep(100000);		/* 0.1s to make sure log is written */
+	exit(ecode);
 }
 
 /* panic: we're still alive but shouldn't */
@@ -138,12 +104,12 @@ static void panic(void)
 	exit(1);
 }
 
-static void mnt_off()
+static void mnt_off(void)
 {
 	FILE *fp;
 	struct mntent *mnt;
 
-	fp = setmntent(MNTTAB, "r");
+	fp = setmntent(_PATH_MNTTAB, "r");
 	/* in some rare cases fp might be NULL so be careful */
 	while (fp != NULL && ((mnt = getmntent(fp)) != (struct mntent *)0)) {
 		/* First check if swap */
@@ -198,7 +164,7 @@ static void mnt_off()
 static PROC *plist;
 
 /* get a list of all processes */
-static int readproc()
+static int readproc(void)
 {
 	DIR *dir;
 	struct dirent *d;
@@ -222,10 +188,10 @@ static int readproc()
 			continue;
 
 		/*
-                * Get a PROC struct. If this fails, which is likely if we have an
-                * out-of-memory error, we return gracefully with what we have managed
-                * so hopefully a 2nd call after killing some processes will give us more.
-                */
+		 * Get a PROC struct. If this fails, which is likely if we have an
+		 * out-of-memory error, we return gracefully with what we have managed
+		 * so hopefully a 2nd call after killing some processes will give us more.
+		 */
 		if ((p = (PROC *) calloc(1, sizeof(PROC))) == NULL) {
 			log_message(LOG_ERR, "out of memory");
 			closedir(dir);
@@ -258,31 +224,30 @@ static void killall5(int sig)
 
 	/* Since we ignore all signals, we don't have to worry here. MM */
 	/* Now stop all processes. */
+	suspend_logging();
 	kill(-1, SIGSTOP);
 
 	/* Find out our own 'sid'. */
-	if (readproc() < 0) {
-		kill(-1, SIGCONT);
-		return;
+	if (readproc() == 0) {
+		for (p = plist; p; p = p->next)
+			if (p->pid == daemon_pid) {
+				sid = p->sid;
+				break;
+			}
+		/* Now kill all processes except our session. */
+		for (p = plist; p; p = p->next)
+			if (p->pid != daemon_pid &&	/* Skip our process */
+				p->sid != sid && 	/* Skip our session */
+				p->sid != 0) 		/* Skip any kernel process. */
+					kill(p->pid, sig);
 	}
-	for (p = plist; p; p = p->next)
-		if (p->pid == daemon_pid) {
-			sid = p->sid;
-			break;
-		}
-	/* Now kill all processes except our session. */
-	for (p = plist; p; p = p->next)
-		if (p->pid != daemon_pid &&	/* Skip our process */
-			p->sid != sid && 	/* Skip our session */
-			p->sid != 0) 		/* Skip any kernel process. */
-				kill(p->pid, sig);
-
 	/* And let them continue. */
 	kill(-1, SIGCONT);
+	resume_logging();
 }
 
-/* shut down the system */
-void do_shutdown(int errorcode)
+/* part that tries to shut down the system cleanly */
+static void try_clean_shutdown(int errorcode)
 {
 	int i = 0, fd;
 	char *seedbck = RANDOM_SEED;
@@ -334,7 +299,7 @@ void do_shutdown(int errorcode)
 								"Message from watchdog:\nIt is too hot to keep on working. The system will be halted!\n");
 						else
 							fprintf(ph,
-								"Message from watchdog:\nThe system will be rebooted because of error %d!\n", errorcode);
+								"Message from watchdog:\nThe system will be rebooted because of error %d = '%s'\n", errorcode, wd_strerror(errorcode));
 						if (ferror(ph) != 0) {
 							log_message(LOG_ERR, "cannot send mail (errno = %d)", errno);
 						}
@@ -348,8 +313,6 @@ void do_shutdown(int errorcode)
 		}
 	}
 
-	/* now tell syslog what's happening */
-	log_message(LOG_ALERT, "shutting down the system because of error %d", errorcode);
 	close_logging();
 
 	safe_sleep(10);		/* make sure log is written and mail is send */
@@ -376,15 +339,15 @@ void do_shutdown(int errorcode)
 	kill(1, SIGTSTP);
 
 	/* Kill all other processes. */
-	(void)killall5(SIGTERM);
+	killall5(SIGTERM);
 	safe_sleep(1);
 	/* Do this twice in case we have out-of-memory problems. */
-	(void)killall5(SIGTERM);
-        safe_sleep(4);
-	(void)killall5(SIGKILL);
+	killall5(SIGTERM);
+	safe_sleep(sigterm_delay-1);
+	killall5(SIGKILL);
 	keep_alive();
 	/* Out-of-memory safeguard again. */
-	(void)killall5(SIGKILL);
+	killall5(SIGKILL);
 	keep_alive();
 
 	/* Remove our PID file, as nothing should be capable of starting a 2nd daemon now. */
@@ -439,19 +402,36 @@ void do_shutdown(int errorcode)
 	mnt_off();
 
 	/* umount all partitions */
-	if (setjmp(ret2dog) == 0)
-		umount_all(NULL);
+	umount_all(NULL);
 
 #if 0
 	/* with the more recent version of mount code, this is not needed anymore */
 	/* remount / read-only */
-	if (setjmp(ret2dog) == 0)
+	//if (setjmp(ret2dog) == 0)
 		mount_one(rootfs.mnt_fsname, rootfs.mnt_dir, rootfs.mnt_type,
 			  rootfs.mnt_opts, rootfs.mnt_freq, rootfs.mnt_passno);
 #endif
 
 	/* shut down interfaces (also taken from sysvinit source */
 	ifdown();
+}
+
+/* shut down the system */
+void do_shutdown(int errorcode)
+{
+	/* tell syslog what's happening */
+	log_message(LOG_ALERT, "shutting down the system because of error %d = '%s'", errorcode, wd_strerror(errorcode));
+
+	if(errorcode != ERESET)	{
+		try_clean_shutdown(errorcode);
+	} else {
+		/* We have been asked to hard-reset, make basic attempt at clean filesystem
+		 * but don't try stopping anything, etc, then used device (below) to do reset
+		 * action.
+		 */
+		sync();
+		sleep(1);
+	}
 
 	/* finally reboot */
 	if (errorcode != ETOOHOT) {
@@ -464,13 +444,13 @@ void do_shutdown(int errorcode)
 		reboot(RB_AUTOBOOT);
 	} else {
 		if (temp_poweroff) {
-		        /* Tell system to power off if possible. */
-		        reboot(RB_POWER_OFF);
+			/* Tell system to power off if possible. */
+			reboot(RB_POWER_OFF);
 		} else {
-		        /* Turn on hard reboot, CTRL-ALT-DEL will reboot now. */
-		        reboot(RB_ENABLE_CAD);
-		        /* And perform the `halt' system call. */
-		        reboot(RB_HALT_SYSTEM);
+			/* Turn on hard reboot, CTRL-ALT-DEL will reboot now. */
+			reboot(RB_ENABLE_CAD);
+			/* And perform the `halt' system call. */
+			reboot(RB_HALT_SYSTEM);
 		}
 	}
 
